#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#define PRINTMTX 0

int main (int argc, char ** argv) {

  int n, i, j;
  double stime, etime;
  int size, rank, rlen;
  char myname[MPI_MAX_PROCESSOR_NAME]; 

  MPI_Init (&argc, &argv);

  MPI_Comm_size (MPI_COMM_WORLD, &size);
  
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  MPI_Get_processor_name (myname, &rlen);

  fprintf (stdout, "I am proces %3d of %3d my name is %s (%s)\n", 
    rank, size, myname, argv[0]);

  fflush(stdout);

  if (argc != 2) {
    fprintf (stderr, "usage: %s N\n", argv[0]);
    MPI_Finalize ();
    return EXIT_FAILURE;
  } 

  if ( (n = atoi (argv[1])) < 0) {
    fprintf (stderr, "N must be gt 0\n");
    MPI_Finalize ();
    return EXIT_FAILURE;
  }

  if (size == 2) {
    int to, from;
    double * d, * rd;
    MPI_Status status;

    srand (10);

    if ( (d = (double *) malloc (sizeof (double) * n * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      return EXIT_FAILURE;
    }

    if ( (rd = (double *) malloc (sizeof (double) * n * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      return EXIT_FAILURE;
    }

    if (rank == 0) {
      to = 1;
      from = 1;
      for (i=0; i<n; i++)
        for (j=0; j<n; j++)
          d[(i*n) + j] = 1.0e0 + (double) (1.0e0 * n * 
              (rand() / (RAND_MAX + 1.0e0)));   
    }
    else if (rank == 1) {
      to = 0;
      from = 0;
      for (i=0; i<n; i++)
        for (j=0; j<n; j++)
          d[(i*n) + j] = 0.0e0;
    }

    stime = MPI_Wtime ();
    MPI_Sendrecv(d, n*n, MPI_DOUBLE, to, 10, rd, n*n, 
        MPI_DOUBLE, from, 10, MPI_COMM_WORLD, &status);
    etime = MPI_Wtime ();
    fprintf (stdout, "SendRecv (%d) time 3: %f s\n", rank, etime - stime);

#if PRINTMTX
    fprintf (stdout, " rank: %d \n", rank);
    for (i=0; i<n; i++) {
      for (j=0; j<n; j++) {
        fprintf (stdout, "%10.5f ", rd[(i*n) + j]);
      }
      fprintf (stdout, "\n");
    }
#endif

    free (d);
    free (rd);
  }

  MPI_Finalize ();

  return EXIT_SUCCESS;
}
